import os
import logging
import re
from uuid import uuid1
from random import randrange

from croniter import croniter
from datetime import datetime, date, timedelta

from p2.agent.client import Client as AgentClient
from p2.agent.cronjob import Cronjob
from p2.store.client import Client as StoreClient
from p2.scheduler.Config import Config
from p2.scheduler.Test import P2test
from p2.scheduler.Shifts import Shifts
from p2.scheduler.k8.resources import Resources


class Scheduler:
    """A class to schedule testing campaigns.
    The workflow is roughly divided into 2 steps:
    - first we get a list of what tests (experiments on targets) we would like to run
    - then we try to find a free slot to run them...
    """

    def __init__(self, logger=None, context:str=None):

        try:
            self.log = logger.getLogger() if logger else logging.getLogger(__name__)

            store_url = os.environ.get('P2_STORE_URL')
            agent_url = os.environ.get('P2_AGENT_URL')
            self.log.debug(f"Store URL {store_url}")
            self.log.debug(f"Agent URL {agent_url}")

            self.p2_store_client = StoreClient(store_url)
            if self.p2_store_client is None:
                raise Exception("Error getting a p2.store client object")

            self.p2_agent_client = AgentClient(agent_url)
            if self.p2_agent_client is None:
                raise Exception("Error getting a p2.agent client object")

            self.config = Config()
            self.shifts = Shifts()
            self.to_be_scheduled = list()

            if context:
                self.context = context

        except Exception as err:
            self.log.error(err)
            return None

    def schedule_campaign(self, campaign_id: int):
        """
        - retrieves the campaign with id
        - get targets
        - get experiments
        - see if experiments have to be scheduled...
        """
        self.campaign_id = campaign_id
        self.campaign    = self._get_campaign()
        targets          = self._get_targets()
        experiments      = self._get_experiments()

        context = self.campaign['resources'].get('context', 'minikube')

        # delete old sessions...
        old_sessions = self.p2_store_client.select(
            table='exp_session',
            filter={'start_time|< now() - interval ': '360 days'},
            fetch=['exp_id', 'start_time']
        )
        for session in old_sessions:
            try:
                self.p2_store_client.delete(table='exp_session', key='exp_id',
                                            value=session['exp_id'])
                self.p2_agent_client.delete_cronjob(context=context, id=session['exp_id'])
            except Exception as e:
                self.log.error(f"Deleting old sessions: {e}")

        # Find out if new resources have appeared since last scheduling
        resources_obj = Resources(context=context)
        resources_obj.update()

        self._get_schedule(targets, experiments)
        self._schedule_all()
        return targets

    def _get_schedule(self, targets: list, experiments: list):
        for target in targets:
            for experiment in experiments:
                if  experiment['type'] == 'experiment' and \
                    'profile' in target and \
                    'target' in experiment['profile'] and \
                    target['profile']['name'] == experiment['profile']['target']['name']:
                    self._add_to_schedule(target=target, experiment=experiment)
                elif experiment['type'] == 'template':
                    # if it's a template then any target is valid...
                    self._add_to_schedule( target=target, experiment=experiment)

    def _add_to_schedule(self, target: dict, experiment: dict):
        self.log.info("See if experiment {} on target {} has to be scheduled...".format(
            experiment['filename'], target['name']))

        last_session = self._get_last_session_for_target(experiment=experiment,
                                                         target=target)
        if last_session is not None and 'status' in last_session:
            min_interval = self.min_interval(target, experiment, last_session['status'])
            expired_by = date.today() - date.fromordinal(last_session['scheduled_day'])
            if 'end_time' in last_session and last_session['end_time'] != None:
                expired_by = date.today() - datetime.fromisoformat(last_session['end_time']).date()

            if expired_by.days > min_interval:
                self.to_be_scheduled.append( P2test(
                        campaign = self.campaign,
                        target = target,
                        experiment = experiment,
                        expired = expired_by.days - min_interval))
                return True
        else:
            # no previous run for this experiment on this target, initializing
            # expiration time...
            self.to_be_scheduled.append( P2test(
                campaign = self.campaign,
                target = target,
                experiment = experiment,
                # expired = self.config.target_priority_period[target_priority]
                expired = 100
                ))
            return True
        return False

    def _schedule_all(self):
        """schedule tests sorted by expiretion time"""
        not_scheduled=list()
        to_be_scheduled_sorted = sorted(self.to_be_scheduled,
                                        key=lambda item: item.expired,
                                        reverse=True)
        for test in to_be_scheduled_sorted:
            duration = test.experiment['duration'] if 'duration' in test.experiment else 30
            if (start_at := self.shifts.get_session_shift(duration)) is not None:
                day = date(
                    day=start_at['day'],
                    month=start_at['month'],
                    year=start_at['year']
                )
                self._schedule_this(test, day, start_at)
            else:
                self.log.warning("No days left to schedule experiment {} on target {}".format(
                    test.experiment['filename'], test.target['name']))
                not_scheduled.append(test)
        return not_scheduled


    def _schedule_this(self, test: P2test, day: date, start_at: dict):
        """finds the best slot in the day for this test"""
        """ Gets the max(next_run_time) of the day
            adds that test expected duration and round up to next hour
            check hour is in test shift"""
        """ in case of a template:
            - template profile should hold the template static parameters
            in case these are set in the test campaign, they override the ones
            set in the template. This is handled by the agent (cron) job.
            - session profile should hold the dynamic params
            - Current dynamic params supported:
                - target: target profile
            """
        try:
            mins = "%02d" % start_at['start_minute']
            hour = "%02d" % start_at['start_hour']
            dom  = "%02d" % start_at['day']
            month= "%02d" % start_at['month']
            schedule = f"{mins} {hour} {dom} {month} *"
            cron = croniter(schedule, datetime.now())
            next_run_time = cron.get_next(datetime)
            exp_id = test.target['name'] + '-' + str(uuid1())[:8]
            session = {
                    'filepath': test.experiment['filename'],
                    'campaign_id': self.campaign_id,
                    'context': test.target['context'],
                    'schedule': schedule,
                    'next_run_time': str(next_run_time),
                    'mins': mins,
                    'hour': hour,
                    'dom': dom,
                    'month': month,
                    'dow': '*',
                    'year': start_at['year'],
                    'profile':{'target':test.target},
                    'status': 'loaded'
            }
            exp_id = self.p2_store_client.create_session(exp_id, session)
            if exp_id != None:
                cronjob = Cronjob(
                        id = exp_id,
                        crontab = schedule,
                        filename = test.experiment['filename'],
                        driver = test.experiment['type'],
                        campaign_id = self.campaign_id
                        )
                ret = self.p2_agent_client.create_cronjob(test.target['context'], cronjob)
                if ret != exp_id:
                    raise Exception("Couldn't add cronjob")

                self.p2_store_client.update_session(exp_id, {'status':'scheduled'})

                if test in self.to_be_scheduled:
                    self.to_be_scheduled.remove(test)

                self.log.info("Session {} scheduled".format(exp_id))
            else:
                self.log.error(f"Error scheduling session for experiment {session['filepath']}")
            return exp_id
        except Exception as e:
            self.log.error(f"Couldn't schedule test, exception: {e}")
            return None


    def get_resources(self):
        resources = self.p2_store_client.get_k8s_resources(
                context=self.context,
                status='active',
                kind=self.kind)
        if self.target_pattern is None: return resources

        targets = list()
        #filter only resources that match pattern
        pattern = re.compile(self.target_pattern)
        for resource in resources:
            if pattern.match(resource['name']):
                targets.append(resource)

        # TODO filter if target_profile is set.
        return targets

    def min_interval(self, target, experiment, status):
        """computes min interval (in days ) between tests
            of same type on same target """

        exp_profile = experiment.get('profile',{'weight':1.0})
        target_priority = self._get_target_priority(target)
        weight = exp_profile.get('weight', 1.0)
        if status == 'failed':
            if weight > 0.9: weight = weight - 0.5
        return int(self.config.target_priority_period[target_priority] * weight)

    def _get_target_priority(self, target):
        # target_priority=len(self.config.target_priority_period) or 1
        target_priority = 1
        if target['context'] in self.config.targets_priority:
            target_priority=self.config.targets_priority[target['context']].get(target['name'],
                    len(self.config.target_priority_period))
        return target_priority-1

    def _get_experiments(self):
        return self.p2_store_client.select(
            table='experiments',
            filter=self.campaign['experiments']
        )

    def _get_targets(self):
        return self.p2_store_client.select(
            table='k8s_resources',
            filter=self.campaign['resources']
        )

    def _get_campaign(self):
        campaign = self.p2_store_client.get_campaign_of_id(self.campaign_id)
        if campaign is None:
            raise Exception(f"No campaign with id {id} found")

        if campaign['resources'] is None:
            raise Exception(f"Error, no resources set for campaign {id}")

        if campaign['experiments'] is None:
            raise Exception(f"Error, no experiments set for campaign {id}")
        return campaign

    def _get_last_session_for_target(self, experiment: dict, target: dict):
        """Returns the last session for the passed target, no matter the status """
        sessions = self.p2_store_client.get_sessions(
            filepath=experiment['filename'],
            context=target['context']
            )
        if sessions is None or len(sessions) == 0:
            return None

        valid_sessions = [ s for s in sessions if _valid_session(s, target) ]
        if valid_sessions is None or len(valid_sessions) == 0:
            return None

        for s in valid_sessions:
            s['scheduled_day'] = date(s['year'], int(s['month']),
                                      int(s['dom'])).toordinal()

        sessions_sorted = sorted(valid_sessions, key=lambda session:
                                 session['scheduled_day'],
                                 reverse=True)
        return sessions_sorted[0]

def _valid_session(session: dict, target: dict):
    if (    'profile' not in session
            or session['profile'] == None
            or 'target' not in session['profile']
            or 'name' not in session['profile']['target']
            or session['profile']['target']['name'] != target['name']):
        return False
    else:
        return True
